#property copyright "Read the forum discussion "
#property link      "https://www.mql5.com/en/forum/446275"
#property version   "1.00"
enum execution_mode{
mode_standard=0,//just a program
mode_cpu=1,//openCL_Cpu
mode_gpu=2//openCL_Gpu
};
input execution_mode ExecutionMode=mode_standard;//Execution Mode : 
input string bmpFile="\\Files\\pizzas.bmp";//bmp file location
input int blurMin=0;//blurPulse Min (per side per axis)
input int blurMax=3;//blurPulse Max (per side per axis)
input uint rate=10;//refresh rate (timer) ms

union pixel{
uint  mix;
uchar channels[4];
};
uint original_pixels[],blur_pixels[];
uint width,height;
pixel raw_pixels[];
bool busy=false;
bool loaded=false;
int blurPulse=0,blurStep=0;
string system_tag="TEST_";
long full_pulse_start=0,full_pulse_end=0,full_pulse_time=0;
//open cl handles 
  int OCL_CONTEXTS[],OCL_PROGRAMS[],OCL_KERNELS[],OCL_BUFFERS[];
  bool OCL_EXPECTING_RESULT=false;
int OnInit()
  {
  unload_prep_of_cl(OCL_CONTEXTS,OCL_PROGRAMS,OCL_BUFFERS,OCL_KERNELS);
  OCL_EXPECTING_RESULT=false;
  ObjectsDeleteAll(ChartID(),system_tag);
  ResourceFree("::IMAGE");
  ResourceFree("::BLURRED");
  blurPulse=0;
  blurStep=0;
  full_pulse_start=0;
  full_pulse_end=0;
  full_pulse_time=UINT_MAX;
  loaded=false;
  busy=false;
  EventSetMillisecondTimer(rate);
  return(INIT_SUCCEEDED);
  }
void OnTimer(){
  //if not busy
    if(!busy){
    busy=true;
    
    //if not loaded
      if(!loaded){
      EventKillTimer();
      bool load=ResourceCreate("IMAGE",bmpFile);
      if(load){
      bool create=ResourceReadImage("::IMAGE",original_pixels,width,height);
      if(create){
      ArrayResize(blur_pixels,ArraySize(original_pixels),0);
      ArrayFill(blur_pixels,0,ArraySize(blur_pixels),0);
      ArrayResize(raw_pixels,ArraySize(blur_pixels),0);
      for(int i=0;i<ArraySize(original_pixels);i++){raw_pixels[i].mix=original_pixels[i];}
      //create the displays 
        string name=system_tag+"_BLUR_IMAGE";
        ObjectCreate(ChartID(),name,OBJ_BITMAP_LABEL,0,0,0);
        ObjectSetInteger(ChartID(),name,OBJPROP_XSIZE,width);
        ObjectSetInteger(ChartID(),name,OBJPROP_YSIZE,height);
        ObjectSetInteger(ChartID(),name,OBJPROP_XDISTANCE,10);
        ObjectSetInteger(ChartID(),name,OBJPROP_YDISTANCE,30);
        ObjectSetString(ChartID(),name,OBJPROP_BMPFILE,"::BLURRED");
        if(ExecutionMode==mode_standard){
        loaded=true;
        blurStep=1;
        EventSetMillisecondTimer(rate);
        }else{
        if(prep_open_cl(OCL_CONTEXTS,OCL_PROGRAMS,OCL_BUFFERS,OCL_KERNELS,width,height,0,raw_pixels)){
          loaded=true;
          blurStep=1;
          EventSetMillisecondTimer(rate);
          }else{
          Alert("Cannot setup openCL");
          ExpertRemove();
          }
        }
      }else{
      Alert("Cannot read image #"+IntegerToString(GetLastError()));
      ExpertRemove();
      }
      }else{
      Alert("Cannot load image "+bmpFile);
      ExpertRemove();
      }
      }
      //if loaded
      else if(loaded){
     
      if(ExecutionMode==mode_standard){
      pulse();
      square_blur_image(raw_pixels,width,height,blurPulse,blur_pixels);
      ResourceCreate("BLURRED",blur_pixels,width,height,0,0,width,COLOR_FORMAT_ARGB_NORMALIZE);
      ChartRedraw();
      }else{
      /*
      with open cl need to also know if i'm expecting a response ! 
      and 
      if i'm expecting one and i get a completion then i send the next one !
      */
      if(OCL_EXPECTING_RESULT){
        //and we get it 
          ENUM_OPENCL_EXECUTION_STATUS status=CLExecutionStatus(last_element(OCL_KERNELS));
          if(status==CL_COMPLETE){
          ResetLastError();
          if(CLBufferRead(OCL_BUFFERS[3],blur_pixels,0,0,ArraySize(blur_pixels))){
            ResourceCreate("BLURRED",blur_pixels,width,height,0,0,width,COLOR_FORMAT_ARGB_NORMALIZE);
            pulse();
            ChartRedraw();
            OCL_EXPECTING_RESULT=square_blur_image_open_cl(blurPulse);
            }else{
            Print("Cannot read output buffer #"+IntegerToString(GetLastError()));
            ExpertRemove();
            }
          }else{
          if(status==CL_UNKNOWN){ExpertRemove();}
          ChartRedraw();
          }
        }else{
        OCL_EXPECTING_RESULT=square_blur_image_open_cl(blurPulse);
        }
      }
      

      }
    
    busy=false;
    }
  //if not busy ends here 
  }  
void OnDeinit(const int reason)
  {
  unload_prep_of_cl(OCL_CONTEXTS,OCL_PROGRAMS,OCL_BUFFERS,OCL_KERNELS);
  }
void OnTick()
  {
  }

void pulse(){
      
      //stat keeping
      if(blurPulse==blurMin&&blurStep==1){
        full_pulse_start=GetTickCount();
        }
      //
      blurPulse+=blurStep;
      //stat keeping 
      if(blurPulse==blurMin&&blurStep==-1){
        full_pulse_end=GetTickCount();
        full_pulse_time=full_pulse_end-full_pulse_start;
        if(full_pulse_end<full_pulse_start){full_pulse_time=UINT_MAX-full_pulse_start+full_pulse_end;}
        Comment("Full Pulse Cycle(ms) "+IntegerToString(full_pulse_time));
        }
           if(blurPulse==blurMax){blurStep=-1;}
      else if(blurPulse==blurMin){blurStep=1;}
}

void square_blur_image(const pixel &_pixels[],
                       int _width,
                       int _height,
                       int px_per_side,
                       uint &_output[]){
double max_distance=MathSqrt(MathPow(((double)px_per_side+1),2.0)*2);
for(int y=0;y<_height;y++){
   int y_from=y-px_per_side;if(y_from<0){y_from=0;}
   int y_to=y+px_per_side;if(y_to>=_height){y_to=_height-1;}
   
   for(int x=0;x<_width;x++){
      int x_from=x-px_per_side;if(x_from<0){x_from=0;}
      int x_to=x+px_per_side;if(x_to>=_width){x_to=_width-1;}   
      double mix_sum=0.0,red_sum=0.0,green_sum=0.0,blue_sum=0.0;
      
      for(int mix_y=y_from;mix_y<=y_to;mix_y++){
         for(int mix_x=x_from;mix_x<=x_to;mix_x++){
              int mix_co=mix_y*_width+mix_x;
            //get distance 
              double distance=MathSqrt(MathPow((mix_x-x),2.0)+MathPow((mix_y-y),2.0));
            //get mix coefficient 
              double mix_coeff=(max_distance-distance)/max_distance;
              mix_sum+=mix_coeff;
              red_sum+=mix_coeff*((double)raw_pixels[mix_co].channels[2]);
              green_sum+=mix_coeff*((double)raw_pixels[mix_co].channels[1]);
              blue_sum+=mix_coeff*((double)raw_pixels[mix_co].channels[0]);
            }
         }
      red_sum/=mix_sum;if(red_sum>255.0){red_sum=255.0;}
      green_sum/=mix_sum;if(green_sum>255.0){green_sum=255.0;}
      blue_sum/=mix_sum;if(blue_sum>255.0){blue_sum=255.0;}
      
      int co=y*_width+x;
      _output[co]=(255<<24)|(((uchar)MathFloor(red_sum))<<16)|(((uchar)MathFloor(green_sum))<<8)|((uchar)MathFloor(blue_sum));
      
      }
   }
}
//unload open cl
void unload_prep_of_cl(int &context_handles[],
                       int &program_handles[],
                       int &buffer_handles[],
                       int &kernel_handles[]){
for(int i=0;i<ArraySize(buffer_handles);i++){CLBufferFree(buffer_handles[i]);}
for(int i=0;i<ArraySize(kernel_handles);i++){CLKernelFree(kernel_handles[i]);}
for(int i=0;i<ArraySize(program_handles);i++){CLProgramFree(program_handles[i]);}
for(int i=0;i<ArraySize(context_handles);i++){CLContextFree(context_handles[i]);}
}
//open cl 
bool prep_open_cl(int &context_handles[],
                   int &program_handles[],
                   int &buffer_handles[],
                   int &kernel_handles[],
                   int image_size_x,
                   int image_size_y,
                   int px_per_side,
                   pixel &_pixel_channels[]){
bool setup_result=false;
float max_distance=(float)MathSqrt(MathPow(((float)px_per_side+1),(float)2.0)*2.0);  
//first get context 
  ResetLastError();
  if(ExecutionMode==mode_cpu){
    context_handles[expand_primitive_array(context_handles)]=CLContextCreate(CL_USE_CPU_ONLY);
    }
  else if(ExecutionMode==mode_gpu){
    context_handles[expand_primitive_array(context_handles)]=CLContextCreate(CL_USE_GPU_DOUBLE_ONLY);
    }
//if context valid 
  if(context_handles[last_element_index(context_handles)]!=INVALID_HANDLE){
  ResetLastError();
  //source code 
    /*
    the red green blue channel arrays should be constant but i wont touch it for now as there is no mention
    of constant mem on the native cl docs . 
    */
    string blur_source_code="__kernel void square_blur_pixel(__global float *red_channel_array,\r\n"
                                                            "__global float *green_channel_array,\r\n"
                                                            "__global float *blue_channel_array,\r\n"
                                                            "__global uint *pixels_output_array,\r\n"
                                                            "float max_distance,\r\n"
                                                            "int blur_side_size,\r\n"
                                                            "int full_size_x,\r\n"
                                                            "int full_size_y){\r\n"
                                  "int x=get_global_id(0);\r\n"
                                  "int y=get_global_id(1);\r\n"
                                  "if(x>=full_size_x){return;}\r\n"
                                  "if(y>=full_size_y){return;}\r\n"
                                  "int y_from=y-blur_side_size;if(y_from<0){y_from=0;}\r\n"
                                  "int y_to=y+blur_side_size;if(y_to>=full_size_y){y_to=full_size_y-1;}\r\n"
                                  "int x_from=x-blur_side_size;if(x_from<0){x_from=0;}\r\n"
                                  "int x_to=x+blur_side_size;if(x_to>=full_size_x){x_to=full_size_x-1;} \r\n"  
                                  "float mix_sum=0.0,red_sum=0.0,green_sum=0.0,blue_sum=0.0;\r\n"
                                  "for(int mix_y=y_from;mix_y<=y_to;mix_y++){\r\n"
                                  "   for(int mix_x=x_from;mix_x<=x_to;mix_x++){\r\n"
                                  "      int mix_co=mix_y*full_size_x+mix_x;\r\n"
                                  "      float distance=sqrt(pow((float)(mix_x-x),(float)2.0)+pow((float)(mix_y-y),(float)2.0));\r\n"
                                  "      float mix_coeff=(max_distance-distance)/max_distance;\r\n"
                                  "      mix_sum+=mix_coeff;\r\n"
                                  "      red_sum+=mix_coeff*red_channel_array[mix_co];\r\n"
                                  "      green_sum+=mix_coeff*green_channel_array[mix_co];\r\n"
                                  "      blue_sum+=mix_coeff*blue_channel_array[mix_co];\r\n"
                                  "      }\r\n"
                                  "   }\r\n"
                                  "red_sum/=mix_sum;if(red_sum>255.0){red_sum=255.0;}\r\n"
                                  "green_sum/=mix_sum;if(green_sum>255.0){green_sum=255.0;}\r\n"
                                  "blue_sum/=mix_sum;if(blue_sum>255.0){blue_sum=255.0;}    \r\n"
                                  "int co=y*full_size_x+x;\r\n"
                                  "pixels_output_array[co]=(255<<24)|(((uchar)floor(red_sum))<<16)|(((uchar)floor(green_sum))<<8)|((uchar)floor(blue_sum));     \r\n"                                
                                  "}\r\n";
      string build_log="";
      program_handles[expand_primitive_array(program_handles)]=CLProgramCreate(context_handles[last_element_index(context_handles)],blur_source_code,build_log);    
      if(program_handles[last_element_index(program_handles)]!=INVALID_HANDLE){
      ResetLastError();
      //buffers , oof - red channel 
        buffer_handles[expand_primitive_array(buffer_handles)]=CLBufferCreate(last_element(context_handles),image_size_x*image_size_y*4,CL_MEM_READ_ONLY);
        //green
        buffer_handles[expand_primitive_array(buffer_handles)]=CLBufferCreate(last_element(context_handles),image_size_x*image_size_y*4,CL_MEM_READ_ONLY);
        //blue
        buffer_handles[expand_primitive_array(buffer_handles)]=CLBufferCreate(last_element(context_handles),image_size_x*image_size_y*4,CL_MEM_READ_ONLY);
      //output 
        buffer_handles[expand_primitive_array(buffer_handles)]=CLBufferCreate(last_element(context_handles),image_size_x*image_size_y*4,CL_MEM_WRITE_ONLY);
      //check buffers 
        bool buffers_okay=true;
        for(int i=0;i<ArraySize(buffer_handles);i++){
        if(buffer_handles[i]==INVALID_HANDLE){buffers_okay=false;}
        }
        if(buffers_okay){
        ResetLastError();
        kernel_handles[expand_primitive_array(kernel_handles)]=CLKernelCreate(last_element(program_handles),"square_blur_pixel");
        if(last_element(kernel_handles)!=INVALID_HANDLE){
            ResetLastError();
          //setup kernel arguments   
            if(CLSetKernelArgMem(last_element(kernel_handles),0,buffer_handles[0])){
              ResetLastError();
              if(CLSetKernelArgMem(last_element(kernel_handles),1,buffer_handles[1])){
                ResetLastError();
                if(CLSetKernelArgMem(last_element(kernel_handles),2,buffer_handles[2])){
                  ResetLastError();
                  if(CLSetKernelArgMem(last_element(kernel_handles),3,buffer_handles[3])){
                    ResetLastError();
                    if(CLSetKernelArg(last_element(kernel_handles),4,max_distance)){
                      ResetLastError();
                      if(CLSetKernelArg(last_element(kernel_handles),5,px_per_side)){
                        ResetLastError();
                        if(CLSetKernelArg(last_element(kernel_handles),6,image_size_x)){
                          ResetLastError();
                          if(CLSetKernelArg(last_element(kernel_handles),7,image_size_y)){
                            ResetLastError();
                            //fill the constant channels 
                              //red 
                                float a_channel[];
                                ArrayResize(a_channel,ArraySize(_pixel_channels),0);
                                for(int i=0;i<ArraySize(_pixel_channels);i++){a_channel[i]=(float)_pixel_channels[i].channels[2];}
                                uint filled_red=CLBufferWrite(buffer_handles[0],a_channel,0,0,ArraySize(_pixel_channels));
                              //green 
                                for(int i=0;i<ArraySize(_pixel_channels);i++){a_channel[i]=(float)_pixel_channels[i].channels[1];}
                                uint filled_green=CLBufferWrite(buffer_handles[1],a_channel,0,0,ArraySize(_pixel_channels));                              
                              //blue
                                for(int i=0;i<ArraySize(_pixel_channels);i++){a_channel[i]=(float)_pixel_channels[i].channels[0];}
                                uint filled_blue=CLBufferWrite(buffer_handles[2],a_channel,0,0,ArraySize(_pixel_channels));                                
                              if((filled_red+filled_green+filled_blue)==ArraySize(_pixel_channels)*3){
                                ResetLastError();
                                setup_result=true;
                                Print("Setup okay");
                                }else{Alert("OpenCL::Cannot fill constant channels #"+IntegerToString(GetLastError()));}
                            }else{Alert("OpenCL::Cannot create image size y arg for kernel #"+IntegerToString(GetLastError()));}
                          }else{Alert("OpenCL::Cannot create image size x arg for kernel #"+IntegerToString(GetLastError()));}
                        }else{Alert("OpenCL::Cannot create blur_side_size arg for kernel #"+IntegerToString(GetLastError()));}
                      }else{Alert("OpenCL::Cannot create max distance arg for kernel #"+IntegerToString(GetLastError()));}
                    }else{Alert("OpenCL::Cannot create output args for kernel #"+IntegerToString(GetLastError()));}
                  }else{Alert("OpenCL::Cannot create blue channel args for kernel #"+IntegerToString(GetLastError()));}
              }else{Alert("OpenCL::Cannot create green channel args for kernel #"+IntegerToString(GetLastError()));}
            }else{Alert("OpenCL::Cannot create red channel args for kernel #"+IntegerToString(GetLastError()));}
          //setup kernel arguments 
          }else{Alert("OpenCL::Cannot create kernel #"+IntegerToString(GetLastError()));}
        }else{Alert("OpenCL::Cannot create buffers");}
      }else{Alert("OpenCL::cannot create source code #"+IntegerToString(GetLastError())+"\n"+build_log);}
  }else{Alert("OpenCL::Context error #"+IntegerToString(GetLastError()));}

return(setup_result);
}

bool square_blur_image_open_cl(int px_per_side){
bool result=false;
float max_distance=(float)MathSqrt(MathPow(((float)px_per_side+1),(float)2.0)*2.0);  
//set arguments 
  ResetLastError();
  if(CLSetKernelArg(last_element(OCL_KERNELS),4,max_distance)){
  ResetLastError();
  if(CLSetKernelArg(last_element(OCL_KERNELS),5,px_per_side)){
  ResetLastError();
  int offsets[]={0,0};
  int work_size[]={(int)width,(int)height};
  if(CLExecute(last_element(OCL_KERNELS),2,offsets,work_size)){
    result=true;
    }else{Print("Execution failed for calculation #"+IntegerToString(GetLastError()));}
  }else{Print("Cannot set px_per_side arg for calculateion #"+IntegerToString(GetLastError()));}
  }else{Print("Cannot set max distance arg for calculation #"+IntegerToString(GetLastError()));}
return(result);
}

//expand primitive array
template<typename X>
int expand_primitive_array(X &array[]){
ArrayResize(array,ArraySize(array)+1,0);
return(ArraySize(array)-1);
}

template<typename X>
int last_element_index(X &array[]){
return(ArraySize(array)-1);
}

template<typename X>
X last_element(X &array[]){
return(array[ArraySize(array)-1]);
}